"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""

from __future__ import absolute_import
from .helpers import *
from . import frame
from . import rpc
from . import buffer
